# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jinja2
from elasticsearch.helpers import bulk
from elasticsearch_dsl import (
    Completion,
    connections,
    Date,
    Document,
    Integer,
    Q,
    Search,
    Text,
)


DEFAULT_ES_PORT = 9200
DEFAULT_SUGGEST_SIZE = 5


class ZubbiDoc(Document):
    """All documents which are scraped by Zubbi and stored in Elasticsearch."""

    scrape_time = Date(default_timezone="UTC")

    @classmethod
    def outdated_query(cls, timestamp, extra_filter=None):
        extra_filter = extra_filter or []

        return (
            cls.search()
            .query("bool", filter=extra_filter)
            .filter("range", scrape_time={"lt": timestamp})
        )

    # Helper method as mentioned in
    # https://github.com/elastic/elasticsearch-dsl-py/issues/403
    @classmethod
    def bulk_save(cls, docs):
        objects = (d.to_dict(include_meta=True) for d in docs)
        client = connections.get_connection()
        return bulk(client, objects)


class ZuulTenant(ZubbiDoc):
    tenant_name = Text()

    class Index:
        name = "zuul-tenants"


class GitRepo(ZubbiDoc):
    repo_name = Text()

    class Index:
        name = "git-repos"


class Block(ZubbiDoc):
    name_suggest = Completion(
        contexts=[
            {"name": "private", "type": "category", "path": "private"},
            {"name": "tenants", "type": "category", "path": "tenants"},
        ]
    )
    repo = Text(analyzer="whitespace")
    tenants = Text(multi=True, analyzer="whitespace")
    # NOTE (fschmidt): Elasticsearch does not support context suggestion for
    # Boplean fields. As we are using the private flag to filter the auto-
    # completion results, this must be Text.
    private = Text()
    url = Text()
    description = Text(analyzer="whitespace")
    description_html = Text()
    platforms = Text(multi=True, analyzer="whitespace")
    last_updated = Date(default_timezone="UTC")

    @staticmethod
    def suggest_field():
        return "name_suggest"

    @property
    def has_html_description(self):
        return bool(self.description_html)

    @property
    def has_html_changelog(self):
        return bool(self.changelog_html)

    @property
    def description_rendered(self):
        return self._renderable_field(self.description_html, self.description)

    def _renderable_field(self, html, raw):
        if html:
            return jinja2.Markup(html)
        elif raw:
            return jinja2.Markup("<pre>{}</pre>".format(jinja2.escape(raw)))


class AnsibleRole(Block):
    # NOTE (fschmidt): We have to store the name as 'role_name' in the result,
    # so we can use it for aggregation in Elasticsearch later on.
    role_name = Text(analyzer="whitespace")
    changelog = Text(analyzer="whitespace")
    changelog_html = Text()

    class Index:
        name = "ansible-roles"

    @property
    def name(self):
        return self.role_name

    @property
    def changelog_rendered(self):
        return self._renderable_field(self.changelog_html, self.changelog)

    @classmethod
    def bulk_save(cls, docs):
        for doc in docs:
            doc.name_suggest = doc.role_name
        return super().bulk_save(docs)


class ZuulJob(Block):
    # NOTE (fschmidt): We have to store the name as 'job_name' in the result,
    # so we can use it for aggregation in Elasticsearch later on.
    job_name = Text(analyzer="whitespace")
    parent = Text(analyzer="whitespace")
    line_start = Integer()
    line_end = Integer()

    class Index:
        name = "zuul-jobs"

    @property
    def name(self):
        return self.job_name

    def save(self, **kwargs):
        # Make sure that the name suggestion contains the name
        self.name_suggest = self.job_name
        return super().save(**kwargs)


class BlockSearch(Search):
    def __init__(self, index=None, doc_type=None, block_class=None, **kwargs):
        super().__init__(
            index=self._select_index(index, block_class),
            doc_type=self._select_doc_type(doc_type, block_class),
            **kwargs
        )

    @staticmethod
    def _select_index(index, block_class):
        if index:
            return index
        if block_class is not None:
            return block_class._default_index()
        return [AnsibleRole._default_index(), ZuulJob._default_index()]

    @staticmethod
    def _select_doc_type(doc_type, block_class):
        if doc_type:
            return doc_type
        if block_class is not None:
            return [block_class]
        return [AnsibleRole, ZuulJob]

    def search_query(self, query, fields, exact=False, extra_filter=None):
        # Elasticsearch cannot work with a set, only with list
        fields = list(fields)
        if exact:
            search_query = Q("multi_match", query=query, fields=fields)
        else:
            query_string = "*{}*".format(query)
            search_query = Q("query_string", query=query_string, fields=fields)

        extra_filter = extra_filter or []

        return self.query("bool", filter=extra_filter, must=search_query)

    def detail_query(self, block_name, repo, extra_filter=None):
        extra_filter = extra_filter or []
        detail_query = [
            Q("query_string", query=block_name, fields=["job_name", "role_name"]),
            Q("match", repo=repo),
        ]
        return self.query("bool", filter=extra_filter, must=detail_query)

    def suggest_query(
        self, text, name="suggester", size=DEFAULT_SUGGEST_SIZE, contexts=None
    ):
        contexts = contexts or {}
        return self.suggest(
            name,
            text,
            completion={
                "field": Block.suggest_field(),
                "size": size,
                "contexts": contexts,
            },
        )


def role_type(item):
    return isinstance(item, AnsibleRole)


def job_type(item):
    return isinstance(item, ZuulJob)


def block_type(item):
    if role_type(item):
        return "role"
    elif job_type(item):
        return "job"
    raise ValueError("Unsupported type: {}".format(type(item)))


def class_from_block_type(block_type):
    if block_type == "role":
        return AnsibleRole
    elif block_type == "job":
        return ZuulJob
    raise ValueError("No class for block type: {}".format(block_type))


def init_elasticsearch(app):
    init_elasticsearch_con(
        app.config["ES_HOST"],
        app.config.get("ES_USER"),
        app.config.get("ES_PASSWORD"),
        app.config.get("ES_PORT"),
    )

    app.add_template_test(role_type)
    app.add_template_test(job_type)
    app.add_template_filter(block_type)


def init_elasticsearch_con(host, user=None, password=None, port=None):
    http_auth = None
    # Set authentication parameters if available
    if user and password:
        http_auth = (user, password)
    if port is None:
        port = DEFAULT_ES_PORT
    connections.create_connection(host=host, http_auth=http_auth, port=port)

    ZuulJob.init()
    AnsibleRole.init()
    ZuulTenant.init()
    GitRepo.init()



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest


@pytest.mark.parametrize(
    "endpoint, expected",
    [
        ("/", "200 OK"),
        ("/search", "200 OK"),
        ("/how-to", "200 OK"),
        ("/lmzify", "200 OK"),
    ],
)
def test_views_reachable(flask_client, endpoint, expected):
    """Test if all available views in zubbi are reachable."""
    rv = flask_client.get(endpoint)
    assert rv.status == expected


def test_detail_view(flask_client, es_client):

    # Build a simple ES response, containing a minimal role result
    response = {
        "hits": {
            "hits": [
                {
                    "_index": "ansible-roles",
                    "_type": "doc",
                    "_source": {"role_name": "foo", "repo": "repo_name"},
                }
            ]
        }
    }

    es_client.search.return_value = response
    # TODO (fschmidt): Make this work similar to requests.mock
    # where we can mock each request/response inside the context
    rv = flask_client.get("/detail/repo_name/role/foo")
    assert rv.status == "200 OK"
    assert b"<title>Details for role foo - Zubbi</title>" in rv.data


def test_detail_view_unknown_block_type(flask_client):
    rv = flask_client.get("/detail/repo_name/foobar/foo")
    assert rv.status == "400 BAD REQUEST"
    assert b"Unknown block type" in rv.data


def test_auto_complete_view(flask_client, es_client):
    response = {
        "suggest": {"suggester": [{"options": [{"text": "foo"}, {"text": "foobar"}]}]}
    }

    es_client.search.return_value = response
    rv = flask_client.get("/api/search/autocomplete?term=foo")
    # The resulting list is not ordered
    assert set(rv.get_json()) == {"foobar", "foo"}


@pytest.mark.parametrize(
    "params",
    [
        "",  # no parameter at all
        "?term",  # correct parameter, but missing value
        "?foobar=abc",  # wrong parameter
    ],
)
def test_auto_complete_view_error(flask_client, params):
    rv = flask_client.get("/api/search/autocomplete{}".format(params))
    assert rv.status == "404 NOT FOUND"
    assert rv.get_json() == {"error": 404}


def test_webhook_view_get(flask_client):
    # Don't accept get requests
    rv_get = flask_client.get("/api/webhook")
    assert rv_get.status == "405 METHOD NOT ALLOWED"


def test_webhook_view_missing_payload(flask_client):
    rv_post_empty = flask_client.post("/api/webhook")
    assert rv_post_empty.status == "400 BAD REQUEST"
    assert rv_post_empty.get_json() == {
        "error": 400,
        "msg": "Payload is missing or not a valid JSON",
    }


def test_webhook_view_wrong_sha(flask_client):
    # Valid payload, but wrong sha
    rv_post_invalid = flask_client.post(
        "/api/webhook",
        json={"abc": "cde"},
        headers={
            "X-GitHub-Delivery": "foo",
            "X-Hub-Signature": "wrong_sha",
            "x-Github-Event": "installation",
        },
    )
    assert rv_post_invalid.status == "401 UNAUTHORIZED"
    assert rv_post_invalid.get_json() == {
        "error": 401,
        "msg": "Request signature does not match calculated signature.",
    }


def test_webhook_view_correct_sha(flask_client):
    # Valid payload with precalculated sha
    precalculated_signature = "sha1=3e7397e4c518017be42e2a87522ae117edc422c9"

    rv_post_valid = flask_client.post(
        "/api/webhook",
        json={"abc": "cde"},
        headers={
            "X-GitHub-Delivery": "foo",
            "X-Hub-Signature": precalculated_signature,
            "x-Github-Event": "installation",
        },
    )
    assert rv_post_valid.status == "200 OK"
    assert rv_post_valid.get_json() == {"event_processed": True}



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime, timedelta, timezone
from urllib.parse import urljoin

import jwt
import requests


PREVIEW_JSON_ACCEPT = "application/vnd.github.machine-man-preview+json"

LOGGER = logging.getLogger(__name__)


class GitHubConnection:
    def __init__(self, config):
        self.config = config

        self.base_url = urljoin(self.config["GITHUB_URL"], "api/v3")
        self.graphql_url = urljoin(self.config["GITHUB_URL"], "api/graphql")

        self.app_id = None
        self.app_key = None

        self.installation_map = {}
        self.installation_token_cache = {}

    def onLoad(self):
        LOGGER.debug("Authenticating against GitHub")
        self._authenticate()
        self._prime_install_map()

    def _authenticate(self):
        app_id = self.config.get("GITHUB_APP_ID")
        app_key_file = self.config.get("GITHUB_APP_KEY")
        try:
            with open(app_key_file, "r") as f:
                app_key = f.read()
        except IOError:
            LOGGER.error("Failed to open app key file: %s", app_key_file)

        if not app_id and app_key:
            LOGGER.error(
                "You must provide an app_id and an app_key to use "
                "installation based authentication"
            )
            return

        self.app_id = app_id
        self.app_key = app_key

    def _get_app_auth_headers(self):
        """Set the correctt auth headers to authenticate against GitHub."""
        now = datetime.now(timezone.utc)
        expiry = now + timedelta(minutes=5)

        data = {"iat": now, "exp": expiry, "iss": self.app_id}
        app_token = jwt.encode(data, self.app_key, algorithm="RS256").decode("utf-8")

        headers = {
            "Accept": PREVIEW_JSON_ACCEPT,
            "Authorization": "Bearer {}".format(app_token),
        }

        return headers

    def _get_installation_key(
        self, project, user_id=None, install_id=None, reprime=False
    ):
        """Get the auth token for a project or installation id."""
        installation_id = install_id
        if project is not None:
            installation_id = self.installation_map.get(project, {}).get(
                "installation_id"
            )

        if not installation_id:
            if reprime:
                # prime installation map and try again without refreshing
                self._prime_install_map()
                return self._get_installation_key(
                    project, user_id=user_id, install_id=install_id, reprime=False
                )
            LOGGER.debug("No installation ID available for project %s", project)
            return ""

        # Look up the token from cache
        now = datetime.now(timezone.utc)
        token, expiry = self.installation_token_cache.get(installation_id, (None, None))

        # Request new token if the available one is expired or could not be found
        if (not expiry) or (not token) or (now >= expiry):
            LOGGER.debug("Requesting new token for installation %s", installation_id)
            headers = self._get_app_auth_headers()
            url = "{}/installations/{}/access_tokens".format(
                self.base_url, installation_id
            )

            json_data = {"user_id": user_id} if user_id else None

            response = requests.post(url, headers=headers, json=json_data)
            response.raise_for_status()

            data = response.json()

            token = data["token"]
            expiry = datetime.strptime(data["expires_at"], "%Y-%m-%dT%H:%M:%SZ")
            # Update time zone of expiration date to make it comparable with now()
            expiry = expiry.replace(tzinfo=timezone.utc)

            # Assume, that the token expires two minutes earlier, to not
            # get lost during the checkout/scraping?
            expiry -= timedelta(minutes=2)

            self.installation_token_cache[installation_id] = (token, expiry)

        return token

    def _prime_install_map(self):
        """Fetch all installations and look up the ID for each."""
        url = "{}/app/installations".format(self.base_url)
        headers = self._get_app_auth_headers()
        LOGGER.debug("Fetching installations for GitHub app")

        response = requests.get(url, headers=headers)
        response.raise_for_status()

        data = response.json()

        for install in data:
            install_id = install.get("id")
            token = self._get_installation_key(project=None, install_id=install_id)
            headers = {
                "Accept": PREVIEW_JSON_ACCEPT,
                "Authorization": "token {}".format(token),
            }

            url = "{}/installation/repositories?per_page=100".format(self.base_url)
            while url:
                LOGGER.debug("Fetching repos for installation %s", install_id)
                response = requests.get(url, headers=headers)
                response.raise_for_status()
                repos = response.json()

                # Store all projects in the installation map
                for repo in repos.get("repositories", []):
                    # TODO (fschmidt): Store the installation's
                    # permissions (could come in handy for later features)
                    project_name = repo["full_name"]
                    self.installation_map[project_name] = {
                        "installation_id": install_id,
                        "default_branch": repo["default_branch"],
                    }

                # Check if we need to do further page calls
                url = response.links.get("next", {}).get("url")

    @property
    def repos(self):
        return self.installation_map.keys()

    def get_repos_for_installation(self, install_id):
        return [
            k
            for k, v in self.installation_map.items()
            if v["installation_id"] == install_id
        ]



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
from elasticsearch_dsl.connections import connections

import zubbi
from zubbi.scraper.repos import Repository


FIXTURE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "testdata")


def raw_file(filename="repo_files/zuul.d/jobs.yaml"):
    with open(os.path.join(FIXTURE_DIR, filename)) as raw_file:
        content = raw_file.read()
    return content


class DummyRepo(Repository):
    def __init__(self, name):
        self._name = name

    def check_out_file(self, file_path):
        pass

    def list_directory(self, directory_path):
        pass

    def last_changed(self, path):
        pass

    def url_for_path(self, file_path, highlight_start=None, highlight_end=None):
        return "https://github/{}".format(file_path)

    def blame(self, path):
        pass

    @property
    def url(self):
        return "https://github/{}".format(self._name)

    @property
    def private(self):
        return False

    @property
    def name(self):
        return self._name


@pytest.fixture(scope="function")
def repo_data():
    # TODO (fschmidt): Align this with the MockGitHubRepository in test_integration.py
    # Currently, we always have to adapt both data sets as one is used for the scraper
    # and the other one is used for the parser.
    repo = DummyRepo("my/project")

    tenants = {"jobs": ["foo"], "roles": ["foo", "bar"]}

    job_files = {
        "zuul.d/jobs.yaml": {
            "content": raw_file("repo_files/zuul.d/jobs.yaml"),
            "blame": [],
        },
        "zuul.d/no-jobs.yaml": {
            "content": raw_file("repo_files/zuul.d/no-jobs.yaml"),
            "blame": [],
        },
    }

    role_files = {
        "foo": {
            "last_changed": "2018-09-17 15:15:15",
            "readme_file": {
                "path": "roles/foo/README.rst",
                "content": raw_file("repo_files/roles/foo/README.rst"),
            },
            "changelog_file": None,
        },
        "bar": {
            "last_changed": "2018-09-17 15:15:15",
            "readme_file": {
                "path": "roles/bar/README.rst",
                "content": raw_file("repo_files/roles/bar/README.rst"),
            },
            "changelog_file": None,
        },
    }

    return repo, tenants, job_files, role_files


@pytest.fixture(scope="function")
def github_response_installations():
    response = [
        {
            "id": 94,
            "account": {"login": "admin", "id": 13},
            "integration_id": 15,
            "app_id": 15,
            "target_id": 13,
            "target_type": "Organization",
            "permissions": {"contents": "read", "metadata": "read"},
            "events": [],
            "created_at": "2018-06-21T05:53:40Z",
            "updated_at": "2018-06-21T05:53:40Z",
        }
    ]
    return response


@pytest.fixture(scope="function")
def github_response_repositories():
    response = {
        "total_count": 2,
        "repository_selection": "all",
        "repositories": [
            {
                "id": 186,
                "name": "foobar_repo",
                "full_name": "orga/foo_repo",
                "owner": {
                    "login": "orga",
                    "id": 13,
                    "type": "Organization",
                    "site_admin": False,
                },
                "private": True,
                "description": "foo repo in the orga organization",
                "default_branch": "master",
                "permissions": {"admin": False, "push": False, "pull": False},
            },
            {
                "id": 37,
                "name": "cilib",
                "full_name": "orga/bar_repo",
                "owner": {
                    "login": "orga",
                    "id": 13,
                    "type": "Organization",
                    "site_admin": False,
                },
                "private": False,
                "description": "The bar repo in the orga organization",
                "default_branch": "master",
                "permissions": {"admin": False, "push": False, "pull": False},
            },
        ],
    }

    return response


@pytest.fixture(scope="function")
def github_response_access_token():
    # Mock a future expiry date which will always be valid when calling this
    # function
    expires_future = datetime.now(timezone.utc)
    expires_future += timedelta(minutes=10)
    response = {
        "token": "v1.6a01e67ed5a45dff724d99694a325a090ab69419",
        "expires_at": datetime.strftime(expires_future, "%Y-%m-%dT%H:%M:%SZ"),
    }
    return response


@pytest.fixture(scope="function")
def readme_python_code():
    return raw_file("readmes/python-code.rst")


@pytest.fixture(scope="function")
def readme_yaml_code():
    return raw_file("readmes/yaml-code.rst")


@pytest.fixture(scope="function")
def readme_supported_os():
    return raw_file("readmes/supported-os.rst")


@pytest.fixture(scope="function")
def flask_client(monkeypatch):
    # Mock elasticsearch connection method which is called when the app is created
    monkeypatch.setattr(zubbi.models, "init_elasticsearch_con", mock.Mock())

    config = {
        "ES_HOST": "localhost",
        "GITHUB_WEBHOOK_SECRET": "some_secret",
        "ZMQ_PUB_SOCKET_ADDRESS": "tcp://*:5555",
        "ZMP_SUB_SOCKET_ADDRESS": "tcp://localhost:5555",
        "ZMQ_SUB_TIMEOUT": 10000,
    }

    app = zubbi.create_app(config=config)
    app.config["TESTING"] = True
    client = app.test_client()

    yield client


@pytest.fixture(scope="function")
def es_client():
    # NOTE (fschmidt): I took a look on how elasticsearch-dsl does the mocking:
    # https://github.com/elastic/elasticsearch-dsl-py/blob/master/test_elasticsearch_dsl/conftest.py
    client = mock.Mock()
    client.search.return_value = {}
    connections.add_connection("default", client)
    yield client
    connections._conn = {}
    connections._kwargs = {}



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import github3
import requests

from zubbi.scraper.exceptions import CheckoutError
from zubbi.scraper.repos import Repository


LOGGER = logging.getLogger(__name__)

GRAPHQL_BLAME_QUERY = """
query($owner:String!, $repo:String!, $path:String!) {
    repository(owner: $owner, name: $repo) {
    object(expression: "master") {
        ... on Commit {
        blame(path: $path){
            ranges {
            startingLine
            endingLine
            commit {
                committer {
                date
                }
            }
            }
        }
        }
    }
    }
}
"""


class GitHubRepository(Repository):
    def __init__(self, repo_name, gh, gh_con=None):
        self.repo_name = repo_name
        self.gh = gh
        self.gh_con = gh_con
        self._repo = self._get_repo_object()

    def check_out_file(self, file_path):
        try:
            LOGGER.debug("Checking out '%s'", file_path)
            remote_file_content = self._repo.file_contents(file_path)
            if remote_file_content.size is 0:
                raise CheckoutError(file_path, "File is empty.")
            return remote_file_content.decoded.decode("utf-8")
        except github3.exceptions.NotFoundError:
            raise CheckoutError(file_path, "File not found.")

    def list_directory(self, directory_path):
        try:
            LOGGER.debug("Listing contents of '%s' directory", directory_path)
            # We are only interested in the filenames as the content is None
            # anyways. If necessary, we could add the filetype (file, directory,
            # symlink, ...) to the result (value.type) and use this function
            # in a recursive way. However, this makes only sense if we don't know
            # the file structures for jobs and roles which is not the case
            # for this use case.
            remote_directory = self._repo.directory_contents(
                directory_path, return_as=dict
            )
            return remote_directory
        except github3.exceptions.NotFoundError:
            raise CheckoutError(directory_path, "Directory not found.")

    def last_changed(self, path):
        LOGGER.debug("Getting last changes for '%s'", path)
        # We are only interested in the first (newest) commit
        commits = self._repo.commits(path=path, per_page=1)
        commit = next(commits)
        # Get the full commit object
        git_commit = self._repo.git_commit(sha=commit.sha)
        last_changed = git_commit.committer["date"]
        return last_changed

    def blame(self, path):
        LOGGER.debug("Getting blame info for '%s'", path)
        # TODO (fschmidt): When blame is available in GitHub's V4 API, we should
        # switch to this. Until then, we could use the GraphQL to retrieve the
        # necessary information. I'd like to think that the response from GitHub
        # will look the same, so we need to do the parsing and mapping of the
        # response anyway.
        owner, repo = self.repo_name.split("/", 1)

        variables = {"owner": owner, "repo": repo, "path": path}

        token = self.gh_con._get_installation_key(self.repo_name)
        headers = {"Authorization": "bearer {}".format(token)}
        response = requests.post(
            self.gh_con.graphql_url,
            json={"query": GRAPHQL_BLAME_QUERY, "variables": variables},
            headers=headers,
        )

        if response.status_code != 200:
            return []

        blame_json = response.json()

        # Catch error from GraphQL API
        errors = blame_json.get("errors")
        if errors:
            LOGGER.warning(
                "Could not get blame info for %s in '%s'", path, self.repo_name
            )
            for error in errors:
                LOGGER.warning(error["message"])
            return []

        # Flatten the result
        flat_blame = []
        for blame in blame_json["data"]["repository"]["object"]["blame"]["ranges"]:
            flat_blame.append(
                {
                    "start": blame["startingLine"],
                    "end": blame["endingLine"],
                    "date": blame["commit"]["committer"]["date"],
                }
            )
        return flat_blame

    def _get_repo_object(self):
        owner, repo_name = self.repo_name.split("/")
        repo = self.gh.repository(owner=owner, repository=repo_name)
        return repo

    def url_for_path(self, file_path, highlight_start=None, highlight_end=None):
        file_url = self._repo.file_contents(file_path).html_url

        if highlight_start is not None:
            file_url = "{}#L{}".format(file_url, highlight_start)

            if highlight_end is not None:
                file_url = "{}-L{}".format(file_url, highlight_end)

        return file_url

    @property
    def url(self):
        return self._repo.html_url

    @property
    def private(self):
        return self._repo.private

    @property
    def name(self):
        return self.repo_name



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import logging
import os
import sys
from collections import namedtuple
from datetime import datetime, timedelta, timezone

import click
import github3
import zmq
from elasticsearch.exceptions import ConflictError
from elasticsearch_dsl import Q
from flask.config import Config
from tabulate import tabulate

from zubbi import default_settings
from zubbi import ZUBBI_SETTINGS_ENV
from zubbi.models import (
    AnsibleRole,
    GitRepo,
    init_elasticsearch_con,
    ZuulJob,
    ZuulTenant,
)
from zubbi.scraper.connections.github import GitHubConnection
from zubbi.scraper.exceptions import ScraperConfigurationError
from zubbi.scraper.repo_parser import RepoParser
from zubbi.scraper.repos.github import GitHubRepository
from zubbi.scraper.scraper import Scraper
from zubbi.scraper.tenant_parser import TenantParser


LOGGER = logging.getLogger(__name__)

RepoItem = namedtuple("RepoItem", "name scraped provider")


def configure_logger(verbosity):
    # Import root logger to apply the configuration to all module loggers
    from zubbi.scraper import LOGGER

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    log_formatter = logging.Formatter(
        "%(asctime)s %(levelname)s [%(name)s] %(message)s", datefmt="%Y-%m-%dT%H:%M:%SZ"
    )
    console_handler.setFormatter(log_formatter)

    level = getattr(logging, verbosity.upper())
    environment_level = getattr(
        logging, os.environ.get("ZUBBI_VERBOSITY", verbosity).upper()
    )

    LOGGER.setLevel(min(environment_level, level))
    LOGGER.addHandler(console_handler)


# TODO (fschmidt): Move this method to the GitHubConnection class as it already
# has all the necessary parameters.
def _create_github_client(github_url, gh_con, project):
    """Create a github3 client per repo/installation."""
    token = gh_con._get_installation_key(project=project)
    if not token:
        LOGGER.warning(
            "Could not find an authentication token for '%s'. Do you "
            "have access to this repository?",
            project,
        )
        return
    gh = github3.GitHubEnterprise(github_url)
    gh.login(token=token)
    return gh


def update_tenant_configuration(
    tenant_sources_repo, tenant_sources_file, github_url, gh_con, scrape_time
):
    if tenant_sources_repo:
        gh = _create_github_client(github_url, gh_con, tenant_sources_repo)
        if not gh:
            raise ScraperConfigurationError(
                "Cannot load tenant sources from repo '{}'. No access.".format(
                    tenant_sources_repo
                )
            )

        gh_repo = GitHubRepository(tenant_sources_repo, gh)
        tenant_parser = TenantParser(sources_repo=gh_repo, scrape_time=scrape_time)
    else:
        tenant_parser = TenantParser(
            sources_file=tenant_sources_file, scrape_time=scrape_time
        )

    return tenant_parser.parse()


def _initialize_repo_cache(connections):
    LOGGER.info("Initializing repository cache")
    # Initialize Repo Cache
    repo_cache = {}

    # TODO (fschmidt): Once we have more providers, this could be done in a
    # loop, using all listed connections and add the connection's key as
    # provider in the cached_repo
    gh_con = connections["github"]

    # Get all repos from Elasticsearch
    for hit in GitRepo.search().query("match_all").scan():
        # Add 'github' type if they are listed in our github connection
        if hit.repo_name in gh_con.repos:
            repo_type = "github"
        else:
            repo_type = None
        # TODO (fschmidt): Maybe we can use this list as cache for the whole
        # scraper-webhook part.
        # This way, we could reduce the amount of operations needed for GitHub
        # and ElasticSearch
        repo_cache[hit.repo_name] = {
            "scrape_time": hit.scrape_time,
            "provider": repo_type,
        }

    return repo_cache


@click.group(invoke_without_command=True)
@click.option(
    "--verbosity",
    help="Set the active log level",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"]),
)
@click.pass_context
def main(ctx, verbosity):
    configure_logger(verbosity)

    config = Config(root_path=".")
    config.from_object(default_settings)
    config.from_envvar(ZUBBI_SETTINGS_ENV)

    ctx.obj = {"config": config}

    if ctx.invoked_subcommand is None:
        ctx.invoke(scrape)


@main.command(name="list-repos")
@click.pass_context
def list_repos(ctx):
    repos = []
    config = ctx.obj["config"]

    connections = init_connections(config)
    repo_cache = _initialize_repo_cache(connections)

    # Flatten the repo dict and format the scrape_time for console output
    for key, val in repo_cache.items():
        repos.append(
            RepoItem(
                key,
                datetime.strftime(val["scrape_time"], "%Y-%m-%dT%H:%M:%SZ"),
                val["provider"],
            )
        )

    # Sort repos by name (asc) and scrape time (desc)
    repos = sorted(repos, key=lambda x: x.name)
    repos = sorted(repos, key=lambda x: x.scraped, reverse=True)

    # Print the table
    print(
        tabulate(
            repos,
            tablefmt="orgtbl",
            headers=["Repository", "Last scraped at", "Provider"],
        )
    )


@main.command()
@click.option("--full", "-f", help="Scrape all repositories immediately", is_flag=True)
@click.option("--repo", "-r", help="Scrape only the specified repo", multiple=True)
@click.pass_context
def scrape(ctx, full, repo):

    LOGGER.info("Hello, Zubbi!")
    config = ctx.obj["config"]

    tenant_sources_repo = config.get("TENANT_SOURCES_REPO")
    tenant_sources_file = config.get("TENANT_SOURCES_FILE")

    # Fail if both are set or none of both is set
    if (
        not tenant_sources_file
        and not tenant_sources_repo
        or (tenant_sources_file and tenant_sources_repo)
    ):
        raise ScraperConfigurationError(
            "Either one of 'TENANT_SOURCES_REPO' "
            "and 'TENANT_SOURCES_FILE' must be set, "
            "but not both."
        )

    connections = init_connections(config)
    repo_cache = _initialize_repo_cache(connections)

    if full:
        scrape_full(config, connections)
    elif repo:
        scrape_full(config, connections, repos=repo)
    else:
        # Listen to ZMQ messages
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.connect(config["ZMQ_SUB_SOCKET_ADDRESS"])
        socket.setsockopt_string(zmq.SUBSCRIBE, "")
        socket.setsockopt(zmq.RCVTIMEO, config["ZMQ_SUB_TIMEOUT"])

        while True:
            # Check for incoming messages on ZMQ
            LOGGER.debug("Checking for incoming ZMQ messages")
            try:
                event, payload = socket.recv_multipart()
                handle_event(
                    event.decode("utf-8"),
                    json.loads(payload.decode("utf-8")),
                    config,
                    connections,
                    repo_cache,
                )
            except zmq.error.Again:
                # If no message was received until the timeout, ZMQ throws
                # zmq.error.Again: Resource temporarily unavailable
                LOGGER.debug("Did not receive any ZMQ message")

            # Check if a periodic run is necessary
            LOGGER.debug("Checking for outdated repos")
            scrape_outdated(config, connections, repo_cache)


def init_connections(config):
    # Initialize global GitHub connection for GitHub App
    gh_con = GitHubConnection(config)
    gh_con.onLoad()

    # Initialize Elasticsearch connection
    init_elasticsearch_con(
        config["ES_HOST"],
        config.get("ES_USER"),
        config.get("ES_PASSWORD"),
        config.get("ES_PORT"),
    )

    # NOTE (fschmidt): We could use this one to store e.g. the Gerrit connection
    # also in here
    connections = {"github": gh_con}
    return connections


def scrape_outdated(config, connections, repo_cache):
    scrape_interval = config["FORCE_SCRAPE_INTERVAL"]
    repo_list = []
    now = datetime.now(timezone.utc)
    threshold = now - timedelta(hours=scrape_interval)
    # Check the repo cache for entries older than 24 hours
    for key, val in repo_cache.items():
        # TODO We should clean up repos containing 'None' providers some time
        if val["scrape_time"] < threshold and val["provider"] is not None:
            repo_list.append(key)

    if repo_list:
        LOGGER.info(
            "Found repos which weren't scraped for %d hours: %s",
            scrape_interval,
            repo_list,
        )
        scrape_repo_list(repo_list, config, connections, repo_cache)
    else:
        LOGGER.info(
            "Found no repos which weren't scraped for %d hours: %s",
            scrape_interval,
            repo_list,
        )


def scrape_full(config, connections, repos=None):
    gh_con = connections["github"]
    if repos is None:
        repos = list(gh_con.repos)
    # TODO (fschmidt): Instead of a simple string list, we should provide
    # the whole repo_cache entry containing also the provider info
    # (github, gerrit) to simplify the finding of the correct connection
    scrape_repo_list(repos, config, connections)


def scrape_repo_list(
    repo_list, config, connections, repo_cache=None, delete_only=False
):
    # Simplify the usage of a non-existing repo cache
    if repo_cache is None:
        repo_cache = {}

    scrape_time = datetime.now(timezone.utc)
    tenant_sources_repo = config.get("TENANT_SOURCES_REPO")
    tenant_sources_file = config.get("TENANT_SOURCES_FILE")
    github_url = config.get("GITHUB_URL")
    gh_con = connections["github"]

    LOGGER.info(
        "Using scraping time: %s", datetime.strftime(scrape_time, "%Y-%m-%dT%H:%M:%SZ")
    )

    # TODO (fschmidt): This should only be done once during initialization
    # and when a repo or installation changed

    # Update the installation map
    gh_con._prime_install_map()

    if not delete_only:
        # TODO (fschmidt): This should only be done once during initialization,
        # when a repo or installation changed or for a push event to the
        # TENANT_SOURCES_REPO.
        # This would also mean, that the tenant_configuration needs to be kept
        # in memory, e.g. in the TenantScraper itself (something like the prime
        # and reprime of the installations in the GitHub connection)
        # We also need to identify the repos that were added to / removed from
        # the tenant configuration in the push event.

        # Update tenant sources
        repo_map, tenant_list = update_tenant_configuration(
            tenant_sources_repo, tenant_sources_file, github_url, gh_con, scrape_time
        )

        # First, store the tenants in Elasticsearch
        LOGGER.debug(
            "Updating %d tenant definitions in Elasticsearch", len(tenant_list)
        )
        ZuulTenant.bulk_save(tenant_list)

        LOGGER.info("Scraping the following repositories: %s", repo_list)
        es_repos = []
        for repo_name in repo_list:
            # Get the tenants from the repo map. If we get no tenants, we assume
            # that the repo is not part of the tenant config.
            tenants = repo_map.get(repo_name, None)
            if tenants is None:
                LOGGER.warning(
                    "Repo '%s' is not part of our tenant sources. Skip scraping.",
                    repo_name,
                )
                continue

            # Build the data for the repo itself to be stored in Elasticsearch
            uuid = hashlib.sha1(str.encode(repo_name)).hexdigest()
            repo = GitRepo(meta={"id": uuid})
            repo.repo_name = repo_name
            repo.scrape_time = scrape_time
            es_repos.append(repo)

            cached_repo = repo_cache.setdefault(
                repo_name,
                {
                    # As we are in a GitHub event, we can just assume, the provider
                    # is github
                    "provider": "github"
                },
            )

            # Update the scrape time in cache
            cached_repo["scrape_time"] = scrape_time

            # scrape the repo if is part of the tenant config
            scrape_repo(repo_name, tenants, github_url, gh_con, scrape_time)

        # Store the information for all repos we just scraped in Elasticsearch
        LOGGER.debug("Updating %d repo definitions in Elasticsearch", len(es_repos))
        GitRepo.bulk_save(es_repos)
    else:
        LOGGER.info("Deleting the following repositories: %s", repo_list)
        # NOTE (fschmidt): Usually, this should not delete anything we just scraped.
        # Just to be sure, we will execute this only if delete_only is set.
        delete_outdated(
            scrape_time,
            [GitRepo],
            extra_filter=Q({"terms": {"repo_name.keyword": repo_list}}),
        )
        # Delete the repositories from the repo_cache
        for repo_name in repo_list:
            repo_cache.pop(repo_name, None)

    # In both cases we want to delete outdated data.
    # In case of delete_only, this will be everything!
    LOGGER.info("Deleting outdated data for the following repositories: %s", repo_list)
    delete_outdated(
        scrape_time, [AnsibleRole, ZuulJob], extra_filter=Q("terms", repo=repo_list)
    )


def scrape_repo(repo_name, tenants, github_url, gh_con, scrape_time):
    gh = _create_github_client(github_url, gh_con, repo_name)
    if not gh:
        LOGGER.warning("Skipping GitHub repo '%s'", repo_name)
        return

    gh_repo = GitHubRepository(repo_name, gh, gh_con)
    job_files, role_files = Scraper(gh_repo).scrape()

    jobs, roles = RepoParser(
        gh_repo, tenants, job_files, role_files, scrape_time
    ).parse()

    LOGGER.debug("Updating %d job definitions in Elasticsearch", len(jobs))
    ZuulJob.bulk_save(jobs)

    LOGGER.debug("Updating %d role definitions in Elasticsearch", len(roles))
    AnsibleRole.bulk_save(roles)


def delete_outdated(scrape_time, indices, extra_filter=None):
    # Delete all outdated entries in Elasticsearch
    LOGGER.info(
        "Going to delete outdated scraping results which are older than %s",
        datetime.strftime(scrape_time, "%Y-%m-%dT%H:%M:%SZ"),
    )
    for index in indices:
        try:
            deleted_docs = index.outdated_query(scrape_time, extra_filter).delete()
            LOGGER.info(
                "Deleted %d outdated %ss in Elasticsearch",
                deleted_docs.deleted,
                index.__name__,
            )
        except ConflictError:
            LOGGER.info("Deleted 0 outdated %ss in Elasticsearch", index.__name__)


# TODO (fschmidt): Maybe it's worth to move the event_* methods to a GitHubEventHandler
# class or similar. This way, we could encapsulate different events in their respective
# environment (e.g. GitHub, Gerrit, ...)
def handle_event(event, payload, config, connections, repo_cache):
    LOGGER.info("Handling event '%s'", event)
    try:
        # TODO (fschmidt): Maybe we should change this file/module to be a class
        # to get rid of this module lookup
        this_module = sys.modules[__name__]
        method = getattr(this_module, "event_{}".format(event))
    except AttributeError:
        LOGGER.warning(
            "Could not find an appropriate method to handle event '%s'", event
        )
        return

    try:
        # TODO (fschmidt): What about 'repository' events?
        # To get updates for public/private?
        # https://developer.github.com/v3/activity/events/types/#repositoryevent
        method(payload, config, connections, repo_cache)
    except Exception:
        # TODO (fschmidt): Does it make sense to catch an Exception here?
        # Could we catch anything more specific?
        LOGGER.exception("Error while handling event '%s'", event)


def event_installation(payload, config, connections, repo_cache):
    action = payload.get("action")
    installation_id = payload.get("installation", {}).get("id")
    repositories = payload.get("repositories", [])

    if action == "created":
        LOGGER.info("Scraping new installation %d", installation_id)
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repositories]
        # Scrape them
        scrape_repo_list(repo_names, config, connections, repo_cache)

    if action == "deleted":
        LOGGER.info("Deleting data for installation %d", installation_id)
        gh_con = connections["github"]
        # Get repos for this installation from our GitHubConnection as they are
        # not listed in the payload.
        # FIXME (fschmidt): When the repo map is updated between triggering and
        # handling the event, the repo_list can no longer be retrieved from
        # the installation (which is already deleted by then).
        # As a fallback, we could search for all repos in ES which start with
        # the organization (<orga>/*)
        repositories = gh_con.get_repos_for_installation(installation_id)
        if not repositories:
            LOGGER.warning(
                "Could not retrieve repo list for installation %d. "
                "Maybe we don't have access any longer.",
                installation_id,
            )
            LOGGER.warning("Nothing to delete")
            return
        # Delete all data for those repos
        scrape_repo_list(
            repositories, config, connections, repo_cache, delete_only=True
        )

        # TODO (fschmidt): Should we remove them also from the installatino map?


def event_installation_repositories(payload, config, connections, repo_cache):
    installation_id = payload.get("installation", {}).get("id")
    repos_added = payload.get("repositories_added")
    repos_removed = payload.get("repositories_removed")

    project_name = payload.get("repository", {}).get("full_name")
    print(installation_id)
    print(project_name)

    # TODO validate installation id from payload against installation map?
    # If they do not match, we might have missed an installation event and
    # should update our installation map

    # Scrape each added repo
    if repos_added is not None:
        LOGGER.info(
            "Scraping %d new repositories for installation %d",
            len(repos_added),
            installation_id,
        )
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repos_added]
        # Scrape them
        scrape_repo_list(repo_names, config, connections, repo_cache)

    # Just delete the data for these repos
    if repos_removed is not None:
        LOGGER.info(
            "Deleting data from %d repositories for installation %d",
            len(repos_removed),
            installation_id,
        )
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repos_added]
        # Delete all data for those repos
        scrape_repo_list(repo_names, config, connections, repo_cache, delete_only=True)


def event_push(payload, config, connections, repo_cache):
    repo_name = payload.get("repository", {}).get("full_name")
    # installation_id = payload.get('installation', {}).get('id')
    ref = payload.get("ref")
    # old_ref = payload.get('before')
    # new_ref = payload.get('after')
    # commits = payload.get('commits')

    gh_con = connections["github"]

    repo_info = gh_con.installation_map.get(repo_name)
    default_branch = repo_info["default_branch"]

    # TODO validate installation id from payload against installation map?
    # If they do not match, we might have missed an installation event and
    # should update our installation map

    parts = ref.split("/", 2)
    branch = parts[2]
    if branch != default_branch:
        LOGGER.info(
            "Push event contains ref %s, but default branch is %s. "
            "Won't handle event for repo %s.",
            ref,
            default_branch,
            repo_name,
        )
        return

    LOGGER.info("Handling push event for repo %s with ref %s", repo_name, ref)

    scrape_repo_list([repo_name], config, connections, repo_cache)


if __name__ == "__main__":
    main()



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess

import click
from flask.cli import current_app, with_appcontext


@click.command()
@click.argument("dst", type=click.Path(exists=True))
@with_appcontext
def collectstatic(dst):
    for blueprint in current_app.iter_blueprints():
        subprocess.run(["cp", "-rv", blueprint.static_folder, dst])
    subprocess.run(["cp", "-rv", current_app.static_folder, dst])



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

import requests_mock

from zubbi.scraper.connections.github import GitHubConnection


GITHUB_URL = "https://github.example.com"
GITHUB_API_URL = "{}/api/v3".format(GITHUB_URL)

GITHUB_CON_CONFIG = {
    "GITHUB_APP_ID": 1,
    "GITHUB_APP_KEY": "tests/testdata/app_key_file",
    "GITHUB_URL": GITHUB_URL,
}


def test_get_app_auth_headers():
    # Initialize GitHubConnection
    gh_con = GitHubConnection(GITHUB_CON_CONFIG)
    gh_con._authenticate()

    result = gh_con._get_app_auth_headers()

    assert result["Accept"] == "application/vnd.github.machine-man-preview+json"
    assert result["Authorization"].startswith("Bearer ")


def test_get_installation_key(
    github_response_installations,
    github_response_repositories,
    github_response_access_token,
):

    # Initialize GitHubConnection
    gh_con = GitHubConnection(GITHUB_CON_CONFIG)
    gh_con._authenticate()

    with requests_mock.Mocker() as m:
        # Mock necessary GitHub API endpoints
        m.get(
            "{}/app/installations".format(GITHUB_API_URL),
            json=github_response_installations,
        )
        m.get(
            "{}/installation/repositories?per_page=100".format(GITHUB_API_URL),
            json=github_response_repositories,
        )
        m.post(
            "{}/installations/94/access_tokens".format(GITHUB_API_URL),
            json=github_response_access_token,
        )

        # Calling prime_install_map() includes get_installation_key()
        gh_con._prime_install_map()
        token_from_cache, expires_at = gh_con.installation_token_cache[94]

        # Call get_installation_key() for a specific project (should read from
        # the cache)
        token_for_project = gh_con._get_installation_key(project="orga/foo_repo")

    expected_installation_map = {
        "orga/foo_repo": {"default_branch": "master", "installation_id": 94},
        "orga/bar_repo": {"default_branch": "master", "installation_id": 94},
    }

    assert gh_con.installation_map == expected_installation_map
    assert token_from_cache == "v1.6a01e67ed5a45dff724d99694a325a090ab69419"
    assert token_for_project == "v1.6a01e67ed5a45dff724d99694a325a090ab69419"
    assert isinstance(expires_at, datetime)



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import logging

import yaml
from yaml.composer import Composer
from yaml.constructor import Constructor

from zubbi.doc import render_file, render_sphinx, SphinxBuildError
from zubbi.models import AnsibleRole, ZuulJob
from zubbi.utils import last_changed_from_blame_range, urljoin


LOGGER = logging.getLogger(__name__)


class RepoParser:
    def __init__(self, repo, tenants, job_files, role_files, scrape_time):
        self.repo = repo
        self.tenants = tenants
        self.job_files = job_files
        self.role_files = role_files
        self.scrape_time = scrape_time

    def parse(self):
        LOGGER.info("Parsing files in repo '%s'", self.repo)

        repo_jobs = self.parse_job_files()
        repo_roles = self.parse_roles_dir()
        return repo_jobs, repo_roles

    def parse_job_files(self):
        """Check for job definitions in known zuul files."""
        repo_jobs = []
        for rel_job_file_path, job_info in self.job_files.items():
            LOGGER.debug("Checking for job definitions in %s", rel_job_file_path)
            jobs = self.parse_job_definitions(rel_job_file_path, job_info)
            LOGGER.debug("Found %d job definitions in %s", len(jobs), rel_job_file_path)
            repo_jobs.extend(jobs)
        if not repo_jobs:
            LOGGER.info("No job definitions found in repo '%s'", self.repo)
        else:
            LOGGER.info(
                "Found %d job definitions in repo '%s'", len(repo_jobs), self.repo
            )
            # LOGGER.debug(json.dumps(repo_jobs, indent=4))
        return repo_jobs

    def parse_job_definitions(self, file_path, job_info):
        jobs_yaml = yaml.load(job_info["content"], Loader=ZuulSafeLoader)
        jobs = []
        for list_elem in jobs_yaml:
            job_def = list_elem.pop("job", None)
            if job_def:
                job_name = job_def["name"]
                uuid = hashlib.sha1(
                    str.encode("{}{}".format(self.repo, job_name))
                ).hexdigest()
                # Define basic job data structure
                job = ZuulJob(meta={"id": uuid})
                job.job_name = job_name
                job.repo = self.repo.name
                job.tenants = self.tenants["jobs"]
                job.private = self.repo.private
                job.scrape_time = self.scrape_time
                job.line_start = job_def["__line_start__"]
                job.line_end = job_def["__line_end__"]
                job.last_updated = last_changed_from_blame_range(
                    job.line_start, job.line_end, job_info["blame"]
                )
                job.url = self.repo.url_for_path(
                    file_path, job.line_start, job.line_end
                )

                if "description" in job_def:
                    job.description = job_def["description"]
                    try:
                        doc = render_sphinx(job.description)
                        job.description_html = doc["html"]
                        job.platforms = doc["platforms"]
                    except SphinxBuildError as exc:
                        LOGGER.warning(
                            "Description of job '%s' could not be "
                            "converted to HTML: %s",
                            job_name,
                            exc,
                        )

                # TODO (fschmidt): Look up the tenant.default-parent and
                # use this one over 'base' if no parent is defined.
                # If parent is explicitly set to None, we will keep it.
                job.parent = job_def.get("parent", "base")

                jobs.append(job)
        return jobs

    def parse_roles_dir(self):
        repo_roles = []

        for role_name, role_info in self.role_files.items():
            # We will build the role data structure with or without description
            uuid = hashlib.sha1(
                str.encode("{}{}".format(self.repo, role_name))
            ).hexdigest()
            role = AnsibleRole(meta={"id": uuid})
            role.role_name = role_name
            role.repo = self.repo.name
            role.tenants = self.tenants["roles"]
            role.private = self.repo.private
            role.url = urljoin(self.repo.url, "tree/master/roles", role_name)
            role.scrape_time = self.scrape_time
            role.last_updated = role_info["last_changed"]

            readme_file = role_info.get("readme_file")
            if readme_file:
                # Always store the raw description (can be rendered as fallback)
                role.description = readme_file["content"]
                rendered_content = render_file(readme_file)
                if rendered_content:
                    role.description_html = rendered_content.pop("html")
                    # We might have gotten more results from the parsing (like platforms)
                    # Thus, we simply store those return values directly in the role
                    for k, v in rendered_content.items():
                        setattr(role, k, v)

            changelog_file = role_info.get("changelog_file")
            if changelog_file:
                # Always store the raw description (can be rendered as fallback)
                role.changelog = changelog_file["content"]
                rendered_content = render_file(changelog_file)
                if rendered_content:
                    role.changelog_html = rendered_content.pop("html")

            repo_roles.append(role)

        if not repo_roles:
            LOGGER.info("No role definitions found in repo '%s'", self.repo)
        else:
            LOGGER.info(
                "Found %d role definitions in repo '%s'", len(repo_roles), self.repo
            )
            # LOGGER.debug(json.dumps(repo_roles, indent=4))
        return repo_roles


# Source: https://github.com/openstack-infra/zuul-sphinx/commit/3ef1afe17ee74f5420247463652efd71e6f1e406
class ZuulSafeLoader(yaml.SafeLoader):
    def __init__(self, *args, **kwargs):
        super(ZuulSafeLoader, self).__init__(*args, **kwargs)
        self.add_multi_constructor("!encrypted/", self.construct_encrypted)

    @classmethod
    def construct_encrypted(cls, loader, tag_suffix, node):
        return loader.construct_sequence(node)

    def compose_node(self, parent, index):
        line = self.line
        node = Composer.compose_node(self, parent, index)
        # The line number where the previous token has ended (plus empty lines)
        node.__line_start__ = line + 1
        node.__line_end__ = self.line
        return node

    def construct_mapping(self, node, deep=False):
        mapping = Constructor.construct_mapping(self, node, deep=deep)
        mapping["__line_start__"] = node.__line_start__
        mapping["__line_end__"] = node.__line_end__
        return mapping
